import argparse
import importlib
import logging
import sys
import warnings

from ocdskit.exceptions import CommandError
from ocdskit.util import ijson

logger = logging.getLogger('ocdskit')

COMMAND_MODULES = (
    'ocdskit.commands.combine_record_packages',
    'ocdskit.commands.combine_release_packages',
    'ocdskit.commands.compile',
    'ocdskit.commands.detect_format',
    'ocdskit.commands.echo',
    'ocdskit.commands.indent',
    'ocdskit.commands.mapping_sheet',
    'ocdskit.commands.package_records',
    'ocdskit.commands.package_releases',
    'ocdskit.commands.schema_report',
    'ocdskit.commands.schema_strict',
    'ocdskit.commands.set_closed_codelist_enums',
    'ocdskit.commands.split_record_packages',
    'ocdskit.commands.split_release_packages',
    'ocdskit.commands.upgrade',
)


# The arguments are for use in oc4idskit.
def main(description='Open Contracting Data Standard CLI', modules=COMMAND_MODULES, logger=logger):
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--encoding', help='the file encoding')
    parser.add_argument('--ascii', help='print escape sequences instead of UTF-8 characters', action='store_true')
    parser.add_argument('--pretty', help='pretty print output', action='store_true')

    subparsers = parser.add_subparsers(dest='subcommand')

    subcommands = {}

    for module in modules:
        try:
            command = importlib.import_module(module).Command(subparsers)
        except ImportError as e:
            logger.error('exception "%s" prevented loading of %s module', e, module)  # noqa: TRY400 # UX
        else:
            subcommands[command.name] = command

    args = parser.parse_args()

    if args.subcommand:
        command = subcommands[args.subcommand]
        try:
            command.args = args
            try:
                with warnings.catch_warnings():
                    warnings.showwarning = _showwarning
                    command.handle()
            except ijson.common.IncompleteJSONError as e:
                if e.args and isinstance(e.args[0], (bytes, UnicodeDecodeError)):
                    message = e.args[0]
                    if isinstance(e.args[0], bytes):  # YAJL backends
                        message = e.args[0].decode(errors='backslashreplace')
                    _raise_encoding_error(message, args.encoding)
                raise CommandError(f'JSON error: {e}') from e
            except UnicodeDecodeError as e:
                _raise_encoding_error(e, args.encoding)
        except CommandError as e:
            logger.critical(e)
            sys.exit(1)
    else:
        parser.print_help()


def _showwarning(message, category, filename, lineno, file=None, line=None):  # noqa: ARG001
    if file is None:
        file = sys.stderr
    print(message, file=file)


def _raise_encoding_error(e, encoding):
    suggestion = 'utf-8' if encoding and encoding.lower() == 'iso-8859-1' else 'iso-8859-1'
    raise CommandError(f'encoding error: {e}\nTry `--encoding {suggestion}`?')


if __name__ == '__main__':
    main()



import json

import pytest
from ocdsextensionregistry import ProfileBuilder
from ocdsmerge.exceptions import DuplicateIdValueWarning, InconsistentTypeError

from ocdskit.combine import compile_release_packages, merge, package_records
from ocdskit.exceptions import (
    InconsistentVersionError,
    MergeErrorWarning,
    UnknownVersionError,
)
from tests import read

inconsistent = [{
    "ocid": "ocds-213czf-1",
    "date": "2000-01-01T00:00:00Z",
    "integer": 1
}, {
    "ocid": "ocds-213czf-1",
    "date": "2000-01-02T00:00:00Z",
    "integer": {
        "object": 1
    }
}]


def test_package_default_arguments():
    data = [item for filename in ('realdata/record-package-1.json', 'realdata/record-package-2.json')
            for item in json.loads(read(filename))['records']]

    actual = package_records(data)

    assert actual == json.loads(read('realdata/record-package_record-package.json'))


def test_merge_empty():
    compiled_releases = list(merge([]))

    assert compiled_releases == []


def test_merge_with_schema():
    builder = ProfileBuilder('1__1__4', {'additionalContactPoint': 'master'})
    schema = builder.patched_release_schema()

    data = json.loads(read('release-package_additional-contact-points.json'))['releases']
    compiled_release = next(iter(merge(data, schema=schema)))

    assert compiled_release == json.loads(read('compile_extensions.json'))


def test_merge_without_schema():
    data = json.loads(read('release-package_additional-contact-points.json'))['releases']
    compiled_release = next(iter(merge(data)))

    assert compiled_release == json.loads(read('compile_no-extensions.json'))


def test_merge_warning():
    data = json.loads(read('release-package_warning.json'))['releases']

    with pytest.warns(DuplicateIdValueWarning) as records:
        compiled_release = next(iter(merge(data)))

    assert compiled_release == json.loads(read('compile_warning.json'))

    assert [record.message for record in records] == [
        ("ocds-213czf-1: Multiple objects have the `id` value '1' in the `parties` array"),
    ]


def test_merge_unknown_version(capsys, monkeypatch, caplog):
    def data():
        yield json.loads(read('release-package_unknown-version.json'))

    with pytest.raises(UnknownVersionError):
        list(merge(data()))


def test_merge_unknown_version_force_version(capsys, monkeypatch, caplog):
    def data():
        yield json.loads(read('release-package_unknown-version.json'))

    list(merge(data(), force_version='1.1', ignore_version=True))  # no error


def test_merge_version_mismatch():
    def data():
        yield json.loads(read('realdata/release-package_1.1-1.json'))
        yield json.loads(read('realdata/release-package_1.0-1.json'))

    with pytest.raises(InconsistentVersionError):
        list(merge(data()))


def test_merge_version_mismatch_ignore_version():
    def data():
        yield json.loads(read('realdata/release-package_1.1-1.json'))
        yield json.loads(read('realdata/release-package_1.0-1.json'))

    list(merge(data(), ignore_version=True))  # no error


@pytest.mark.parametrize('return_package', [True, False])
def test_merge_inconsistent_type(return_package):
    def data():
        for release in inconsistent:
            yield {'releases': [release]}

    with pytest.raises(InconsistentTypeError):
        list(merge(data(), return_package=return_package))


@pytest.mark.parametrize(("return_package", "expected"), [
    (
        True,
        [
            {
                'uri': '',
                'publisher': {},
                'publishedDate': '',
                'version': '1.1',
                'records': [{'ocid': 'ocds-213czf-1', 'releases': inconsistent}],
            },
        ],
    ),
    (
        False,
        [],
    ),
])
def test_merge_inconsistent_type_convert_exceptions_to_warnings(return_package, expected):
    def data():
        for release in inconsistent:
            yield {'releases': [release]}

    with pytest.warns(MergeErrorWarning) as records:
        output = list(merge(data(), return_package=return_package, convert_exceptions_to_warnings=True))

    assert output == expected
    assert len(records) == 1
    assert str(records[0].message) == "ocds-213czf-1: An earlier release had the literal 1 for /integer, but the current release has an object with a 'object' key"  # noqa: E501


def test_compile_release_packages():
    with pytest.warns(DeprecationWarning) as records:
        compiled_releases = list(compile_release_packages([]))

    assert compiled_releases == []
    assert len(records) == 1
    assert str(records[0].message) == 'compile_release_packages() is deprecated. Use merge() instead.'



from __future__ import annotations

import warnings

from ocdsextensionregistry import ProfileBuilder
from ocdsmerge import Merger
from ocdsmerge.util import get_release_schema_url

from ocdskit.exceptions import MissingRecordsWarning, MissingReleasesWarning
from ocdskit.packager import Packager
from ocdskit.util import (
    _empty_record_package,
    _empty_release_package,
    _remove_empty_optional_metadata,
    _resolve_metadata,
    _update_package_metadata,
    get_ocds_patch_tag,
)

DEFAULT_VERSION = '1.1'  # fields might be deprecated


def _package(key, items, uri, publisher, published_date, version, extensions=None):
    if publisher is None:
        publisher = {}
    publisher.setdefault('name', '')

    output = {
        'uri': uri,
        'publisher': publisher,
        'publishedDate': published_date,
        'version': version,
    }

    if extensions:
        output['extensions'] = extensions

    output[key] = items

    return output


def package_records(records, uri='', publisher=None, published_date='', version=DEFAULT_VERSION, extensions=None):
    """
    Wraps records in a record package.

    :param list records: a list of records
    :param str uri: the record package's ``uri``
    :param dict publisher: the record package's ``publisher``
    :param str published_date: the record package's ``publishedDate``
    :param str version: the record package's ``version``
    :param list extensions: the record package's ``extensions``
    """
    return _package('records', records, uri, publisher, published_date, version, extensions)


def package_releases(releases, uri='', publisher=None, published_date='', version=DEFAULT_VERSION, extensions=None):
    """
    Wraps releases in a release package.

    :param list releases: a list of releases
    :param str uri: the release package's ``uri``
    :param dict publisher: the release package's ``publisher``
    :param str published_date: the release package's ``publishedDate``
    :param str version: the release package's ``version``
    :param list extensions: the release package's ``extensions``
    """
    return _package('releases', releases, uri, publisher, published_date, version, extensions)


def combine_record_packages(packages, uri='', publisher=None, published_date='', version=DEFAULT_VERSION):
    """
    Collects the packages and records from the record packages into one record package.

    Warns ``~ocdskit.exceptions.MissingRecordsWarning`` if the "records" field is missing from a record package.

    :param packages: an iterable of record packages
    :param str uri: the record package's ``uri``
    :param dict publisher: the record package's ``publisher``
    :param str published_date: the record package's ``publishedDate``
    :param str version: the record package's ``version``
    """
    # See options for not buffering all inputs into memory: https://github.com/open-contracting/ocdskit/issues/119
    output = _empty_record_package(uri, publisher, published_date, version)
    output['packages'] = {}

    for i, package in enumerate(packages):
        _update_package_metadata(output, package)
        if 'records' in package:
            output['records'].extend(package['records'])
        else:
            warnings.warn(
                f'item {i} has no "records" field (check that it is a record package)',
                category=MissingRecordsWarning,
            )
        if 'packages' in package:
            output['packages'].update(dict.fromkeys(package['packages']))

    if publisher:
        output['publisher'] = publisher

    _resolve_metadata(output, 'packages')
    _resolve_metadata(output, 'extensions')
    _remove_empty_optional_metadata(output)

    return output


def combine_release_packages(packages, uri='', publisher=None, published_date='', version=DEFAULT_VERSION):
    """
    Collects the releases from the release packages into one release package.

    Warns ``~ocdskit.exceptions.MissingReleasesWarning`` if the "releases" field is missing from a release package.

    :param packages: an iterable of release packages
    :param str uri: the release package's ``uri``
    :param dict publisher: the release package's ``publisher``
    :param str published_date: the release package's ``publishedDate``
    :param str version: the release package's ``version``
    """
    # See options for not buffering all inputs into memory: https://github.com/open-contracting/ocdskit/issues/119
    output = _empty_release_package(uri, publisher, published_date, version)

    for i, package in enumerate(packages):
        _update_package_metadata(output, package)
        if 'releases' in package:
            output['releases'].extend(package['releases'])
        else:
            warnings.warn(
                f'item {i} has no "releases" field (check that it is a release package)',
                category=MissingReleasesWarning,
            )

    if publisher:
        output['publisher'] = publisher

    _resolve_metadata(output, 'extensions')
    _remove_empty_optional_metadata(output)

    return output


def merge(
    data,
    uri: str = '',
    publisher: dict | None = None,
    published_date: str = '',
    version: str = DEFAULT_VERSION,
    schema: dict | None = None,
    *,
    return_versioned_release: bool = False,
    return_package: bool = False,
    use_linked_releases: bool = False,
    streaming: bool = False,
    force_version: str | None = None,
    ignore_version: bool = False,
    convert_exceptions_to_warnings: bool = False,
):
    """
    Merges release packages and individual releases.

    By default, yields compiled releases. If ``return_versioned_release`` is ``True``, yields versioned releases. If
    ``return_package`` is ``True``, wraps the compiled releases (and versioned releases if ``return_versioned_release``
    is ``True``) in a record package.

    If ``return_package`` is set and ``publisher`` isn't set, the output record package will have the same publisher as
    the last input release package.

    :param data: an iterable of release packages and individual releases
    :param uri: if ``return_package`` is ``True``, the record package's ``uri``
    :param publisher: if ``return_package`` is ``True``, the record package's ``publisher``
    :param published_date: if ``return_package`` is ``True``, the record package's ``publishedDate``
    :param version: if ``return_package`` is ``True``, the record package's ``version``
    :param schema: the URL, path or dict of the patched release schema to use
    :param return_package: wrap the compiled releases in a record package
    :param use_linked_releases: if ``return_package`` is ``True``, use linked releases instead of full releases,
        if the input is a release package
    :param return_versioned_release: if ``return_package`` is ``True``, include versioned releases in the record
        package; otherwise, yield versioned releases instead of compiled releases
    :param streaming: if ``return_package`` is ``True``, set the package's records to a generator (this only works
        if the calling code exhausts the generator before ``merge`` returns)
    :param force_version: version to use instead of the version of the first release package or individual release
    :param ignore_version: do not raise an error if the versions are inconsistent across items to merge
    :param convert_exceptions_to_warnings: whether to convert inconsistent type errors from OCDS Merge to warnings
    :raises InconsistentVersionError: if the versions are inconsistent across items to merge
    :raises MissingOcidKeyError: if the release is missing an ``ocid`` field
    :raises UnknownVersionError: if the OCDS version is not recognized
    """
    with Packager(force_version=force_version) as packager:
        packager.add(data, ignore_version=ignore_version)

        if not schema and packager.version:
            tag = get_ocds_patch_tag(packager.version)
            if packager.package['extensions']:
                # `extensions` is an insertion-ordered dict at this point.
                builder = ProfileBuilder(tag, list(packager.package['extensions']))
                schema = builder.patched_release_schema()
            else:
                schema = get_release_schema_url(tag)

        merger = Merger(schema)

        if return_package:
            packager.package['uri'] = uri
            packager.package['publishedDate'] = published_date
            packager.package['version'] = version
            if publisher:
                packager.package['publisher'] = publisher

            yield from packager.output_package(
                merger,
                return_versioned_release=return_versioned_release,
                use_linked_releases=use_linked_releases,
                streaming=streaming,
                convert_exceptions_to_warnings=convert_exceptions_to_warnings,
            )
        else:
            yield from packager.output_releases(
                merger,
                return_versioned_release=return_versioned_release,
                convert_exceptions_to_warnings=convert_exceptions_to_warnings,
            )


def compile_release_packages(*args, **kwargs):
    warnings.warn('compile_release_packages() is deprecated. Use merge() instead.', DeprecationWarning, stacklevel=2)
    yield from merge(*args, **kwargs)
